# Group 3
## Members
* John Nguyen
* Ibrahem Merhi
* Khaled Rami Omar

## Link to repository
https://gitlab.au.dk/au-ece-sw3hal/e2024/Group-3.git

# Øvelse: Hot-pluggable Device Driver (Raspberry Pi)

Med udgangspunkt i kodeskabelonen til en platform driver, `plat_drv`, skal du undersøge hvordan hot-plugging virker i Linux kontekst og bagefter udvide skabelonen til en GPIO driver.

## Formål

Formålet med denne øvelse er at opnå erfaring med at anvende Linux' device model. Herunder at opnå forståelse for anvendelsen af Hot- og Cold-plugging og hvordan dette implementeres i en device driver. 

Følgende Rubric beskriver forskellig opfyldelse af øvelsens mål:

| Mål                                       | Mangelfuld opfyldelse af mål                              | Delvis opfyldelse af mål 2                              | Fuld opfyldelse af mål                     |
|-------------------------------------------|---------------------------------------|---------------------------------------|------------------------------|
| Undersøge Hot-plugging                    | <ul><li> Driverens init/probe er implementeret </li><li> Driver kan indsættes og device treee loades.</li><li>Der ses IKKE en besked fra probe() i dmesg.</li></ul> | <ul><li> Driverens init/probe er implementeret </li><li> Driver kan indsættes og device treee loades </li><li>  Der ses en besked fra probe() i dmesg. </li><li> Der er vist output fra test</li></ul> | <ul><li> Driverens init/probe er implementeret </li><li> Driver kan indsættes og device treee loades </li><li>  Der ses en besked fra probe() i dmesg. </li><li> Der er vist output fra test </li><li> Driver/device binding er diskuteret </li></ul> | 
| Implementere gpio hot-plug driver         | <ul><li> Der er implementeret basal oprettelse og konfiguration af gpios i probe </li><li> Device oprettes i probe().</li></ul> | <ul><li> Der er implementeret basal oprettelse og konfiguration af gpios i probe </li><li> Device oprettes i probe(). </li><li> Read og write funktioner er opdateret til at kunne tilgå de korrekte gpios.</li></ul> | <ul><li> Der er implementeret basal oprettelse og konfiguration af gpios i probe </li><li> Device oprettes i probe(). </li><li> Read og write funktioner er opdateret til at kunne tilgå de korrekte gpios. </li><li> Der er lavet korrekt cleanup i hhv. exit og remove. </li></ul>|
| Anvende Device Tree                      | <ul><li> Overlay (.dts) er opdateret med GPIOs | <ul><li> Overlay er opdateret med GPIOs </li><li> overlay data bliver læst i probe() </li></ul> | <ul><li> Overlay er opdateret med GPIOs </li><li> overlay data bliver læst i probe() </li><li> gpios oprettes korrekt i /dev </li><li> gpios fungerer korrekt </li><li> tests er dokumentet </li></ul>|

## Delopgave 1: Hot-Plugging af Platform Driver og Device
Målet med denne delopgave er at du opnår forståelse af principperne bag Linux' device model, i dette tilfælde hvordan hot-plug virker og hvilke funktioner som kaldes hvor og hvornår.


### a) Kopier din GPIO driver fra øvelse 4 til en ny folder, kopiér desuden Makefile og plat_drv_overlay.dts fra repositoriet hertil. Omdøb din driver .c fil til plat_drv.c. Byg koden vha. den tilhørende Makefile. Bemærk at resultatet er to filer: kernemodulet plat_drv.ko og et device tree blob overlay, plat_drv.dtbo. Den sidste beskriver hardwaren og er noget vi vil dykke ned i senere.

![Tilføjlse af de tre filer](X1a1.png)

makefilen køres og vi får filerne: "plat_drv.dtbo""plat_drv.ko":

![.ko og .dtbo filerne bygget](X1a2.png)


### b) Tilføj hot-plugging funktionalitet ved at oprette probe() og remove() funktionerne i driveren (frie funktioner ligesom init(), read() mm.), lav en enkelt printk() i hver af dem ("hello from probe\n"...). Tilføj hot-plugging identificering ved at oprette of_device_id og platform_driver struct'ene. Sæt compatible = "au-ece, plat_drv", som er navnet brugt i device tree filen, plat_drv-overlay.dts. Registrer platform driveren i init (som det sidste) vha. platform_driver_register(). Byg og kopiér .ko og .dtbo filer til target.

Der tilføres følgene elementer i plat_drv.c. Der vises kun Det tilføjet element og ikke hele funktionen.
probe():
```c
static int plat_drv_probe(struct platform_device *pdev) {
    {...}
    pr_info("hello from probe\n");    
    {...}
}
```
probe() kaldes automatisk, når en driver finder en enhed, der matcher driverens compatible specifikation fra Device Tree.

Remove():
```c
static int plat_drv_remove(struct platform_device *pdev) {
    pr_info("goodbye from remove\n");
    {...}
}
```
remove() kaldes, når driveren fjernes.

of_device_id strukturen:
```c
static const struct of_device_id led_of_match[] = {
    { .compatible = "au-ece,plat_drv" },  // Compatible name used in Device Tree
};
MODULE_DEVICE_TABLE(of, led_of_match);
```
Dette array bruges til at matche driveren med en compatible værdi fra Device Tree.

platform_driver strukturen:
```c
static struct platform_driver led_platform_driver = {
    .probe = plat_drv_probe,
    .remove = plat_drv_remove,
    .driver = {
        .name = "plat_drv",
        .of_match_table = led_of_match,
        .owner = THIS_MODULE,
    },
};
```
platform_driver-strukturen samler driverens funktioner som probe og remove samt opsætter driverens of_match_table for Device Tree-kompatibilitet.

Registrer platformdriveren i init():
```c
static int __init mygpio_init(void) {
    {...}
    return platform_driver_register(&led_platform_driver);
}
```
Hvilken headerfil skal du inkludere for at benytte platform_driver_register? For at benytte platform_driver_register skal du inkludere følgende headerfil:
```c
#include <linux/platform_device.h>
```
Makefilen køres igen, for at bygge programmen med de nye emplementering. Derefter kopires og overføres .ko og .dtbo til RPi'en:
![Overføresel af .ko og .dtbo til target](X1b1.png)

### c) Du skal nu registrere device't i Linux og indsætte driveren. Du registrerer device't ved at indlæse .dtbo filen på følgende vis:

```c
dtoverlay -d . plat_drv.dtbo
```

Dette svarer til når man laver hot-plugging med USB. Det overordnede RaspberryPi ZW device tree udvides herved til også at indeholde vores device. -d . gør at den loader .dtbo filen fra vores lokale sti, ellers ville den forsøge at loade den fra /boot/overlays. Du skal loade driveren på sædvanlig vis.

Der køres også:
```c
insmod plat_drv.ko
```
Der fåes beskeden, der er skrevet ind i funktion, ved at skrive "dmesg" ("Hello from probe"):

![Besked fra probe funktionen](X1c1.png)

Er driveren tildelt major nummer (se /proc/devices)?  
Når "dmesg" køres ses at major er 239, derfor burde man kunne se i device nummer 239 devicet for led:

![Devises](X1c2.png)

Det er er tildelt en major nummer, som findes i /proc/devices.

Er der lavet en Sys FS klasse (se /sys/class)?

![Class](X1c3.png)

Der er lavet en led klasse, som ses under /sys/class.

Hvilke metoder er blevet kaldt i device driveren (plat_drv.c)? 
probe() kaldes ved tilføjelse af overlayet, og remove() kaldes, når overlayet fjernes.

Hvad sker der hvis vi fjerner overlayet igen (dtoverlay -R plat_drv)? 
Vis resultatet af dmesg og forklar sammenhængen til det observerede med koden i plat_drv.c og sekvensdiagrammet i lektionens slides. Benyt gerne dit eget sekvensdiagram med rigtige funktionsnavne.

![Fjernelse af overlayet](X1c4.png)

Beskeden i remove() vises. Det gøres fordi, ved at lavelse af overlay, laves en device og en class, og det er probe der gør det, mens ved fjernelse af overlay, er det remove der stræder i kræft og hermed vil beskeden.


### d) I probe() skal nu oprettes et device vha device_create(). I remove() tilføjes device_destroy(). Byg og test på target.
Der tilføjes det nødvændige til de to funktioner:
probe():
```c
static int plat_drv_probe(struct platform_device *pdev) {
    {...}
    device_create(led_class, NULL, devno, NULL, "plat_drv");
    {...}
}
```
Remove():
```c
static int plat_drv_remove(struct platform_device *pdev) {
    {...}
    device_destroy(led_class, devno);
    {...}
}
```
Koden overføres til target of testes:

![Køresel af kode](X1d1.png)

Er der oprettet Sys FS devices (se /sys/class/<my_class>/)? 

![Klasse led](X1d2.png)

Der ses at der er oprettet en klasse "led_class"

Er der oprettet noder (/dev)? 

![Node](X1d3.png)

Ja der oprettes noder

virker read/write metoderne?

![read write](X1d4.png)

Der kan godt tændes for lampen vha at ændre på plat_drv værdien. Herud over kan man aflæse i tidligere billeder at beskederne fra de forskellige funktioner printes ud efter "dmesg".



## Delopgave 2: Konstruktion af egen hot-pluggable GPIO device driver
Nu hvor du har prøvet at anvende den device modellens overordnede mekanik, er det på tide at du anvender den til at forbedre din GPIO driver. Din driver skal automatisk oprette 2 noder under /dev/ når device tree filen loades og fjerne dem igen, når den fjernes.
Før vi går løs kan man med fordel lave en lille struct i driveren til at gemme info om GPIO port numre og retninger, samt lave et array af denne som vi senere kan bruge til at mappe minor numre og GPIOs, ex: GPIO12 er input og skal styres vha minor 0:
```c
struct gpio_dev {
  int no;   // GPIO number
  int dir; // 0: in, 1: out
};
static struct gpio_dev gpio_devs[2] = {{12,0}, {21, 1}};
static int gpios_len = 2;
```
Vi har tidligere anvendt request_gpio(), gpio_direction...() i init(). Vi ønsker at allokere GPIO'er dynamisk og bruge GPIO numre og retninger givet i device tree. request_gpio() mm skal derfor flyttes til probe(), da denne kaldes når der er detekteret et gyldigt device (ex. plat_drv i device tree).

### a) Flyt request_gpio(), gpio_direction...() kaldende til probe(). Loop igennem gpios_devs arrayet og brug informationerne til at sætte parametre (GPIO nummer) og funktionskald (gpio_direction_in / out). Lav et kald til device_create() for hver GPIO og brug loop variablen til at sætte minor numret med. Dvs hver GPIO får hver sit minor nummer:
```
[request_gpio() -loop antal]
[gpio_set_direction() -loop antal]
[opret devices med device_create -loop antal]
```
Byg og test på target.

Request og direction() er allerede lavet inde i probe, derfor behøves de ikke at flyttes, men der skal ændres på koden fra del 1. Følgene ændringer/tilføjelser er lavet:

Struct:
```c
#define NUM_GPIOS 2  // Antal GPIO'er

struct gpio_dev {
    int no;   // GPIO nummer
    int dir;  // 0: input, 1: output
};

static struct gpio_dev gpio_devs[NUM_GPIOS] = {
    {12, 0},  // GPIO12, input
    {21, 1}   // GPIO21, output
};
```
Der tilføjes struct GPIO_dev som der bliver spurgt om i opgaven.

Probe():
```c
static int plat_drv_probe(struct platform_device *pdev) {
    int err, i;
    dev_t curr_devno;

    // Allocate Major/Minor numbers
    err = alloc_chrdev_region(&devno, first_minor, max_devices, "led-driver");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        return err;
    }
    pr_info("LED driver got Major %i\n", MAJOR(devno));

    // Create class
    led_class = class_create(THIS_MODULE, "led_class");
    if (IS_ERR(led_class)) {
        pr_err("Failed to create class\n");
        err = PTR_ERR(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
    }

    // Initialize and add cdev
    cdev_init(&led_cdev, &led_fops);
    err = cdev_add(&led_cdev, devno, NUM_GPIOS);  // Registrer to devices
    if (err < 0) {
        pr_err("Failed to add cdev\n");
        class_destroy(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
    }

    // Loop over gpio_devs array og anmod om GPIO'er samt sæt deres retning
    for (i = 0; i < NUM_GPIOS; i++) {
        struct gpio_dev *dev = &gpio_devs[i];
        
        // Anmod om GPIO
        err = gpio_request(dev->no, "gpio_dev");
        if (err) {
            pr_err("Failed to request GPIO %d\n", dev->no);
            goto cleanup_gpio;
        }

        // Sæt GPIO retning
        if (dev->dir == 1) {  // Hvis det er output
            err = gpio_direction_output(dev->no, 0);  // Initialiser til 0
        } else {  // Hvis det er input
            err = gpio_direction_input(dev->no);
        }
        if (err) {
            pr_err("Failed to set GPIO direction for GPIO %d\n", dev->no);
            gpio_free(dev->no);
            goto cleanup_gpio;
        }

        // Opret device node for hver GPIO
        curr_devno = MKDEV(MAJOR(devno), MINOR(devno) + i);
        device_create(led_class, NULL, curr_devno, NULL, "plat_drv%d", i);
        pr_info("Created device for GPIO %d with minor %d\n", dev->no, MINOR(curr_devno));
    }

    pr_info("hello from probe\n");
    return 0;

    cleanup_gpio:
        while (i--) {
            gpio_free(gpio_devs[i].no);
            device_destroy(led_class, MKDEV(MAJOR(devno), MINOR(devno) + i));
        }
        cdev_del(&led_cdev);
        class_destroy(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
}
```
I den opdaterede plat_drv_probe-funktion er der tilføjet logik til korrekt initialisering af en LED-driver. Funktionen allokerer først major og minor numre og opretter en klasse for driveren. Derefter initialiserer den en karakterenhed og registrerer flere GPIO-enheder. Hver GPIO anmodes, og dens retning indstilles til enten input eller output. Hvis der opstår fejl under disse trin, udføres der oprydning ved at frigive GPIO'er og destruere enheder.

remove():
```c
static int plat_drv_remove(struct platform_device *pdev) {
    int i;
    // Fjern devices, cdev og class
    for (i = 0; i < NUM_GPIOS; i++) {
        gpio_free(gpio_devs[i].no);
        device_destroy(led_class, MKDEV(MAJOR(devno), MINOR(devno) + i));
    }
    {...}
}
```
I den opdaterede plat_drv_remove-funktion er der tilføjet logik til at frigive GPIO-enheder og destruere tilknyttede enheder. Funktionen itererer gennem GPIO'enhederne, frigiver dem med gpio_free og destruerer enhederne med device_destroy.

Test:
"dmesg" køres efter overføresel af filerne til target:

![X2a1](X2a1.png)

Der ses at der laves to devices og at probe kører som den skal.

Oprettes der Sys FS devices? 

![X2a2](X2a2.png)

Ja det gøres der for hvrt element i array'et.

Oprettes der devices i /dev/? 

![X2a3](X2a3.png)

Ja man kan se de to også inde i /dev.

Hvilke GPIO'er virker read/write funktionerne på
GPIO 12 og 21, siden det er dem der er indtastet i array'et. 

### b) Vi skal have gjort read/write funktionerne generelle, så de duer på de GPIO'er vi har i vores struct. Dvs. opdatér read/write funktioner til at anvende gpio_devs struct'en. Husk at vi har adgang til minor nummeret i read/write vha:
```c
size_t plat_drv_read(struct file *filep, char __user *ubuf, size_t count, loff_t *f_pos)
{
  int minor;
  minor = iminor(filep->f_inode);

  [gpio_get_value(gpio_devs[.......]) ]
}
```
Byg og test at det virker på de GPIO'er som du har angivet i gpios_devs. Dokumentér med output af dmesg og resultater af dine tests og forklar hvordan de afspejler din implementering.

Write():
```c
ssize_t mygpio_write(struct file *filep, const char __user *ubuf, size_t count, loff_t *f_pos) {
    int len, value;
    char kbuf[12];
    int minor = MINOR(filep->f_inode->i_rdev);  // Få minor nummer
    struct gpio_dev *dev;

    // Tjek om minor nummeret er gyldigt
    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];  // Få den korrekte gpio_dev struct baseret på minor nummeret

    // Kontroller, om GPIO'en er sat til output
    if (dev->dir != 1) {
        pr_err("GPIO %d is not configured as output\n", dev->no);
        return -EINVAL;
    }

    len = count < sizeof(kbuf) ? count : sizeof(kbuf) - 1;

    if (copy_from_user(kbuf, ubuf, len)) {
        return -EFAULT;
    }

    kbuf[len] = '\0';  // Null-terminér strengen

    if (kstrtoint(kbuf, 10, &value)) {
        pr_err("Error converting string to int\n");
        return -EINVAL;
    }

    gpio_set_value(dev->no, value);  // Sæt GPIO-værdien (output)

    pr_info("Wrote %i to GPIO %d\n", value, dev->no);
    *f_pos += len;
    return len;
}
```
I mygpio_write bruges minor nummeret til at finde den specifikke GPIO-enhed i gpio_devs arrayet, så vi skriver til den korrekte GPIO. Funktionen sikrer, at GPIO’en er konfigureret som output, da kun output-GPIO’er kan få en værdi skrevet til dem. Hvis ikke, returneres fejlen -EINVAL. Funktionen læser derefter data fra ubuf, konverterer det til 0 eller 1, og skriver denne værdi til GPIO’en, hvilket sætter spændingen høj eller lav afhængigt af input.

Read():
```c
ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;
    int minor = MINOR(filep->f_inode->i_rdev);  // Få minor nummer
    struct gpio_dev *dev;

    // Tjek om minor nummeret er gyldigt
    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];  // Få den korrekte gpio_dev struct baseret på minor nummeret

    // Kontroller, om GPIO'en er sat til input
    if (dev->dir != 0) {
        pr_err("GPIO %d is not configured as input\n", dev->no);
        return -EINVAL;
    }

    value = gpio_get_value(dev->no);  // Læs GPIO-værdien (input)
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos > 0) {
        return 0;  // EOF
    }

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len;
    return len;
}
```
I mygpio_read bruges minor nummeret til at finde og læse fra den korrekte GPIO-enhed. Funktionen tjekker, at GPIO’en er konfigureret som input; ellers returneres fejlen -EINVAL. Derefter læser den GPIO’ens værdi med gpio_get_value(), som returnerer 0 eller 1 afhængigt af spændingen. Denne værdi konverteres til en streng og kopieres til buf, så brugeren kan aflæse GPIO’ens status.

Test:

![X2b1](X2b1.png)

![X2b2](X2b2.png)

Man kan ved hjælp af echo tænde og slukke for en lampe. fx "echo 1 > /dev/plat_drv1".

![X2b3](X2b3.png)

Denne opdatering gør read og write funktionerne generelle, så de kan bruges på de GPIO'er, der er defineret i gpio_devs structen. Ved at bruge minor nummeret vælges der automatisk den korrekte GPIO, og implementeringen håndterer dynamisk tildeling af input og output baseret på konfigurationen i gpio_devs. Dette gør koden fleksibel og nem at udvide med flere GPIO'er, hvis det ønskes.


## Delopgave 3: Anvendelse af Device Tree i driver

Vi vil nu udvide vores device tree fil, .dts, til at indeholde en beskrivelse af hvilke GPIO'er vores driver skal bruge. Man kan på denne måde ændre GPIO'erne uden at skulle ændre driveren efterfølgende. Du bør have set lektionen og eksempler på how device tree overlays virker. Der er nogle WIKI sider og værktøjer til at oprette Device Tree Overlays som du kan bruge som hjælp.

### a) Opdater plat_drv-overlay.dts til at inkludere GPIO pins. Brug de korrekte gpio numre og hhv '0' for input og '1' for output. I nedenstående eksempel er anvendt GPIO 55 som input og GPIO 56 som output:
```
/dts-v1/;
/plugin/;
/ {
    compatible = "brcm,bcm2835";

    fragment@0 {
        target-path = "/"; /* Add new virtual device to base path */
        __overlay__ {
            mydevt: mydevt {/* devtree label:instance name */
                /* Label to match in driver */
                compatible = "my_device_name";

                /* Configure GPIO module */
                /* <ressource pinno dir pinno dir pinno dir ....*/
                gpios = <&gpio 55 0>, <&gpio 56 1>; 

                /* Custom Property */
                /* <32-bit int value*/
                mydevt-custom = <0x12345678>;
                status = "okay"; /* Device is enabled */
            };
        };
    };
};
```
Dts filen bygges sammen med kernemodulet, når du bruger Makefile'n med make. Bemærk at source filen plat_drv-overlay.dts bygges til filen plat_drv.dtbo. Omdøbningen sker i Makefile'n og skyldes navnekonventioner som anvendes ifb. med Device Tree Overlays.

Der ændres på plat_drv-overlay.dts, så den følger skabelonen, der er udgivet, men med de rigtige GPIO numre samt in og out osv.:
```
/dts-v1/;
/plugin/;

/ {
    compatible = "brcm,bcm2835";

    fragment@0 {
        target-path = "/";
        __overlay__ {
            plat_drv: plat_drv@0 {
                compatible = "au-ece, plat_drv";
                status = "okay";

                /* Definer GPIO'er til driveren */
                gpios = <&gpio 12 0>, /* GPIO 12 som input */
                        <&gpio 21 1>; /* GPIO 21 som output */

                /* Tilføj evt. en brugerdefineret værdi */
                mydevt-custom = <0x12345678>;
            };
        };
    };
};
```
I ovenstående:
- gpio 12 0 betyder, at GPIO 12 er konfigureret som input (0 betyder input).
- gpio 21 1 betyder, at GPIO 21 er konfigureret som output (1 betyder output).

Når .dts-filen bygges, vil Makefile generere en .dtbo-fil (Device Tree Overlay Binary) ved at omdøbe .dtb-filen til .dtbo som beskrevet i  Makefilen.


### b) Anvend Device Tree parametre i device driveren

Anvend of (Open Firmware) funktionerne i probe() til at udlæse GPIO numre og retning (in/out) fra Device Tree. Værdierne kan med fordel tilskrives elementerne i gpio_devs struct'en fra tidligere og efterfølgende bruges til initialisering af GPIO'er og oprettelse af devices (som I sådan set allerede er implementeret...).

Vigtige funktioner fra of_gpio.h:
```c 
 static inline int of_gpio_count(struct device_node *np)
 static inline int of_get_gpio(struct device_node *np, int index)
 static inline int of_get_gpio_flags(struct device_node *np, int index, enum of_gpio_flags *flags)
```
of_get_gpio() returnerer GPIO nummer eller en værdi <0 ved fejl. of_get_gpio_flags() returnerer flagets værdi by reference via flags parameteren. Returværdien er det tilhørende GPIO nummer eller en værdi <0 ved fejl.

I driveren skal du opdatere probe() noget i stil med:

```c
#include <linux/err.h>
#include <linux/of_gpio.h>
#include <linux/platform_device.h>

/* GPIO Device Data Struct */
struct gpio_dev {
  int gpio;
  int flag;
};

static struct gpio_dev gpio_devs[255];
int gpio_devs_cnt = 0;

static int plat_drv_probe(struct platform_device *pdev)
{
  int err = 0;
  struct device *dev = &pdev->dev; // Device ptr derived from current platform_device
  struct device_node *np = dev->of_node; // Device tree node ptr
  enum of_gpio_flags flag;
  int gpios_in_dt = 0;

  /* Retrieve number of GPIOs */
  [hent antal af gpioer (of_gpio_count)]

  /* Loop through gpios in Device Tree */
  for (int i=0;i<gpios_in_dt ; i++) {
    /* hent gpio nummer (of_get_gpio) og skriv i struct */ 
    /* Hent gpio flag, dvs retning (of_get_gpio_flags)  og skriv i struct */
  }

  for (int i=0;i<gpios_in_dt ; i++) {
    /* request_gpio[gpio_devs[i].no]  */ 
    /* Sæt gpio direction iht .dir i struct */
    /* Opret devices med device_create() */
    printk(KERN_ALERT "Using GPIO[%i], flag:%i on major:%i, minor:%i\n",
             gpio_devs[gpio_devs_cnt].gpio, gpio_devs[gpio_devs_cnt].flag,
             MY_DRIVERS_major_number, gt_devs_cnt);
   }
}
```
Du kan finde inspiration i spi-oc-tiny.c driveren som findes på jeres VMware image under /home/stud/sources/rpi-5.4.51/drivers/spi/spi-oc-tiny.c. Se funktionerne tiny_spi_probe() og tiny_spi_of_probe() (som kaldes i tiny_spi_probe())

```c
static int plat_drv_probe(struct platform_device *pdev) {
    int err, i;
    dev_t curr_devno;
    struct device *dev = &pdev->dev;
    struct device_node *np = dev->of_node;
    enum of_gpio_flags flag;

    // Hent antal GPIO'er fra Device Tree
    gpios_len = of_gpio_count(np);
    if (gpios_len <= 0) {}

    pr_info("Found %d GPIOs in Device Tree\n", gpios_len);

    // Loop gennem GPIO'er i Device Tree og hent konfiguration
    for (i = 0; i < gpios_len; i++) {}

    // Opret karakterenheder og GPIO-konfigurationer baseret på Device Tree-data
    for (i = 0; i < gpios_len; i++) {
        // Anmod om GPIO
        // Sæt GPIO-retning
        // Opret device node for hver GPIO
    }

    pr_info("Driver successfully probed with GPIO configuration from Device Tree\n");
    return 0;

    cleanup_gpio:
}
```
Ændringerne i koden gør platformdriveren mere dynamisk og fleksibel ved at hente GPIO-konfigurationer fra Device Tree i stedet for at være afhængig af faste værdier. NUM_GPIOS er blevet erstattet af en variabel, gpios_len, der automatisk tilpasses efter det antal GPIO'er, der er defineret i Device Tree. Desuden er der tilføjet en of_match_table, hvilket gør det muligt for driveren at matche enheder baseret på deres compatible-strenge. Samlet set forbedrer disse ændringer driverens modularitet og genanvendelighed på tværs af forskellige hardwarekonfigurationer.

## Delopgave 4: Automatisk load af Device Tree Overlay og driver modul

For at loade device og driver automatisk under boot skal du gøre følgende 5 ting (ex: mit_modul.ko, mit_overlay.dtbo):

1. Kopiere mit_modul.ko til /lib/modules/<linux-version>/
2. Registrere modulet med depmod -a
3. Tilføje en entry i /etc/modules_load.d/ for at loade det automatisk under boot
4. Kopiere mit_overlay.dtbo til /boot/overlays/
5. Lave entry i /boot/config.txt for at loade device definition under boot
--------------------
1. Kernemoduler ligger som standard under /lib/modules/den_version_af_linux_som_er_i_brug/ (Version kan udlæses med uname -r). Vil du gøre dit kernemodul tilgængelig for systemet, skal du kopiere din .ko fil hertil.
2. Kør depmod -a, som undersøger modul dependencies og laver et load script. Modulet vil herefter kunne loades med modprobe mit_modul.
3. Linux skanner filerne under /etc/modules-load.d/ under upstart for at finde ud af hvilke moduler som skal loades. Du skal derfor oprette en fil her, som indeholder navnet på modulet som skal loades:
```sh
echo mit_modul /etc/modules-load.d/mit_modul.conf
```
4. Kopiér dit overlay, mit_overlay.dtbo, til /boot/overlays på din Rpi. Dette er default søgesti for overlays.
5. Lav en backup af din config.txt fil ved at kopiere den til et nyt navn:
```sh
cp /boot/config.txt /boot/config.txt.backup
```
6. Lav en entry i /boot/config.txt med navnet på overlayet, herved vil Linux søge efter det i /boot/overlays og indlæse det under boot. Brug en editor, f.eks. nano til at editere /boot/config.txt:
```sh
## Angiv overlay (uden .dtbo og ingen mellemrum omkring =)
dtoverlay=mit_overlay
```
BRUG IKKE echo bla bla /boot/config.txt da det overskriver den eksisterende config.txt og din RPI vil ikke længere kunne boote. Du skal I så fald have hentet en config.txt ned fra en medstuderende og kopieret den ind på kortet eller kopiere din backup over i din config.txt. Du skal i begge tilfælde sætte dit SD kort i en kortlæser og mounte dem i Linux. config.txt ligger på FAT partitionen.

Nu er både kernemodul og device tree blob registreret og disse vil automatisk blive loadet når Linux booter. Hvis der er match i "compatible" navnene i hhv. device tree og driver, vil der blive oprettet devices under /dev/ og entries i sysfs. Du kan checke om dit overlay bruger det rigtige navn ved at udlæse /sys/firmware/devicetree/instans_navn_i_mit_overlay/compatible.

Test:
De følgene trin køres og laves:

![X41](X41.png)

Efter genstart tjekkes der om der bliver oprettet noder i /dev.

![X42](X42.png)