# Group 3
# Members
* John Nguyen
* Ibrahem Merhi
* Khaled Rami Omar

# Link to repository
https://gitlab.au.dk/au-ece-sw3hal/e2024/Group-3.git


# Øvelse: LDD med Timers og Attributes (Rpi)

## Formål

Formålet med denne øvelse opnår erfaring med at SysFS attributter og
timers ved er at opdatere en GPIO device driver således at den anvender
SysFS attributter til at styre en ny attribute som lader en GPIO toggle.

Du forventes med øvelsen at opfylde følgende mål:
| **Mål**                      | **Delmål 1**                        | **Delmål 2**                           | **Delmål 3**                      |
|------------------------------|-------------------------------------|----------------------------------------|----------------------------------|
| Oprettelse af attributter    | show/store funktioner               | Deklarere dem og gøre dem til default attributter | Bygge til- og teste på target   |
| Lave attribute funktionalitet | Implementer timer state             | Implementer timer delay                | Test med printk                  |
| Lave timer funktion          | Implementer timer callback           | Test på board                          | Mål timers præcision og jitter   |

## Godkendelse

I bestemmer selv om I vil aflevere øvelse 7 eller 8. Deadline er som
angivet i lektionsplanen på Brightspace.\
Såfremt I aflverer denne, skal Besvarelsen indeholde relevante
kodeudsnit som der henvises til i forklarende tekst, tests og diskussion
af resultater heraf. Komplet kode inklusiv Makefiles lægges i
repository.

## Forberedelse

-   Du skal have læst de relevante sider i kapitel 4+5 i Essential Linux
    Device Drivers
-   Se på source koden i
    \~/sources/rpi-5.4.83/drivers/leds/led-class.c + leds-gpio.c
-   Se detaljer omkring "Reading/Writing Attribute Data" i sysfs.txt. Se
    linje 165 og frem og eksempler linje 229 og frem
-   Kopier platform driveren fra øvelse 6 til en ny folder, vi arbejder
    videre med denne.
-   Du skal have en Raspberry Pi Zero, samt en ASE fHAT med komponenter
    monteret som vist på figuren herunder:

![](../images/ase_fhat_board_irq.png)

## Øvelsen

Du skal i øvelsen udvide din platform driver til at kunne toggle GPIO
output portene ved hjælp af en timer funktion. Toggle delay og hvorvidt
gpio'en toggler (toggle_state) skal kunne sættes og udlæses ved hjælp af
attributes.

Den overordnede struktur for driveren er således (**eksempel!!!
attribute, *my_led_state***):

```C
#include <linux/gpio.h>
...
static ssize_t my_led_state_show(struct device *dev, struct device_attribute *attr, char *buf) 
{ ... }

static ssize_t my_led_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) 
{  ... }
...
DEVICE_ATTR_RW(my_led_state);
static struct attribute *my_led_attrs[] = { ... }
ATTRIBUTE_GROUPS(my_led); // Creates led_groups 
...
static int __init my_platform_driver_init(void)
{
    ...
    gpio_class = class_create(THIS_MODULE, "my_led");
    gpio_class->dev_groups = my_led_groups;

    err = platform_driver_register(&my_led_platform_driver);
    ...
}
```

Bemærk ifht. eksemplerne i bogen, at funktionen *class_device_create()*
er erstattet med *device_create()* og I derfor ikke skal spekulere på
manipulation af kobjekter. **Bemærk desuden at *my_led_state* kun er et
eksempel, ikke navnet på den attribute som I selv skal lave!!**

### Oprettelse af device attributer

Opret show + store funktioner for attributten *gpio_toggle_state*.
Funktionerne har prototyperne:

```C
/* Global Variable */
static int toggle_state;
/* Sysfs "read" method prototype */
static ssize_t gpio_toggle_state_show(struct device *dev, struct device_attribute *attr, char *buf)
/* Sysfs "write" method prototype */
static ssize_t gpio_toggle_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
```

Indtil videre skal *_store()* funktionen alene gemme en værdi i en global variabel, mens *_show()* udlæser den. Lav gerne en printk i de to funktioner til debugging, husk "\\n" i tekststrengen.

Attributten skal nu deklareres og angives som en default attribute for vores class. Benyt nedenstående makroer og struct til at deklarere variable.

Fra *device.h*
[\~/sources/rpi-5.4.83/include/linux/device.h](http://git.kernel.org/?p=linux/kernel/git/stable/linux-stable.git;a=blob;f=include/linux/device.h);
):

```C
// Example of creating an r/w attribute "state" for the class "led" and adding it to the attribute group "led"

DEVICE_ATTR_RW(led_state); // Creates dev_attr_led_state 

static struct attribute *led_attrs[] = { 
    &dev_attr_led_state.attr, 
    // More attributes could go in here, 
    NULL, 
    } ;

ATTRIBUTE_GROUPS(led); // Creates led_groups 
```

Herefter skal du i driverens init funktion sætte dev_groups på den nyoprettede class pointer. Det er vigtigt at du gør det umiddelbart efter *class_create()* og inden *platform_register()*. Kaldet til *platform_register()* gør at *probe()* kaldes med det samme, hvis der allerede er loaded et device tree.

Metoden der er brugt i bogen giver mulighed for at lave specifikke attributter for et givet device. Det står så derfor i kontrast til det I laver her, hvor alle "devices" af samme "class" har/får samme adfærd ("nedarver").

#### Løsning:

Der startes med at tilføje, de nøsvændige funktioner ind i platform driver fra lektion 6:
```c

static ssize_t mygpio_state_show(struct device *dev, struct device_attribute *attr, char *buf) 
{ ... }

static ssize_t mygpio_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) 
{  ... }
...
DEVICE_ATTR_RW(mygpio_state);
static struct attribute *mygpio_attrs[] = { ... }
ATTRIBUTE_GROUPS(mygpio); // Creates led_groups 
...
static int __init my_platform_driver_init(void)
{
    ...
    gpio_class = class_create(THIS_MODULE, "my_led");
    gpio_class->dev_groups = my_led_groups;

    err = platform_driver_register(&my_led_platform_driver);
    ...
}
```
Derefter tilføjes de givne globale variabler:
```c
/* Global Variable */
static int toggle_state;
/* Sysfs "read" method prototype */
static ssize_t gpio_toggle_state_show(struct device *dev, struct device_attribute *attr, char *buf)
/* Sysfs "write" method prototype */
static ssize_t gpio_toggle_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
```

Derefter skal my_led_attrs[] udfyldes, med de givne kode:
```c
static struct attribute *my_led_attrs[] = { 
    &dev_attr_led_state.attr, 
    // More attributes could go in here, 
    NULL,
 }
```

Herefter skrives debug beskederne i _show og _stor funktionerne:
```c
/* Sysfs "read" method prototype */
static ssize_t gpio_toggle_state_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    printk(KERN_INFO "gpio_toggle_state_show called, state: %d\n", gpio_toggle_state); // Debug besked
}
/* Sysfs "write" method prototype */
static ssize_t gpio_toggle_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
    printk(KERN_INFO "gpio_toggle_state_store called, state: %d\n", gpio_toggle_state); // Debug besked
}
```

Der sættes herefter dev_groups ind det korrekte sted og der videreudvikles på _init funktionen:
```c
static int __init mygpio_init(void) {
    int err;

    pr_info("Initializing GPIO Platform Driver\n");

    // Alloker chrdev region
    err = alloc_chrdev_region(&devno, 0, NUM_GPIOS, "plat_drv");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        return err;
    }

    // Opret klassen
    gpio_class = class_create(THIS_MODULE, "my_led");
    if (IS_ERR(gpio_class)) {
        unregister_chrdev_region(devno, NUM_GPIOS);
        return PTR_ERR(gpio_class);
    }

    // Sæt dev_groups til klassen
    gpio_class->dev_groups = my_led_groups;

    // Opret cdev og tilføj den
    cdev_init(&led_cdev, &led_fops);
    led_cdev.owner = THIS_MODULE;
    err = cdev_add(&led_cdev, devno, NUM_GPIOS);
    if (err) {
        pr_err("Failed to add cdev\n");
        class_destroy(gpio_class);
        unregister_chrdev_region(devno, NUM_GPIOS);
        return err;
    }

    // Registrer platformdriveren
    err = platform_driver_register(&my_led_platform_driver);
    if (err) {
        pr_err("Failed to register platform driver\n");
        cdev_del(&led_cdev);
        class_destroy(gpio_class);
        unregister_chrdev_region(devno, NUM_GPIOS);
        return err;
    }

    return 0;
}
```

#### Test af atttributter på target

Indsæt modul og load devicetree. Bemærk at der i /sys/ under dit/dine nye device(s) er kommet en attribute. Skriv og læs på attributten og check i dmesg at det lykkedes.

Der startes med at registrerer device't ved at indlæse .dtbo filen:
```
dtoverlay -d . plat_drv.dtbo
```

Derefter indsættes modulet for .ko filen:
```
insmod plat_drv.ko
```
Der køres "dmesg" for at se om device tree er kørt:

![Dmesg_X1](Dmesg_X1.png)

Der testes om klassen for plat_drv er oprettet:

![Klasse_X1](Klasse_X1.png)

Der bude være oprettet to atributter under klassen (drv0 og drv1):

![Attributter_X1](Attributter_X1.png)

Inde i hvert atribut burde der være en en gpio_state fil, som man kan ændre værdien på til at udskrive en besked med "dmesg".

![gpio_state_X1](gpio_state_X1.png)

Nu testes der om funktionerne store() og show() fungere som de skal vha. cat og echo komand, der aflæses på dmesg om de fungere:
```
echo 1 > gpio_state
cat > gpio_state
```

![skrive_X1](skrive_X1.png)

Ud fra billederne af testen finder vi ud af at show() og store() fungere og kører som den skal, samt printer forventede beskeder når "dmesg" køres.


### Implementering af toggler attributter

Vi går nu videre med den allerede oprettet attribute og udvider med funktionalitet og en attribute mere:

* gpio_toggle_state - Sætter/udlæser vores GPIO toggle tilstand (1=toggling / 0=not toggling), dvs den skal oprette/starte- og slette timeren alt efter om værdien er 0 eller 1.
* gpio_toggle_delay - Sætter/udlæser 1/toggle hastigheden, dvs den skal sætte en variabel som indeholder hvor meget der skal lægges til expiration time i timerfunktionen.

Der skal bruges følgende funktioner (fra linux/timer.h):

* "timer_setup(timer, callback, flags)":https://elixir.bootlin.com/linux/latest/source/include/linux/timer.h#L126\
* "extern int mod_timer(struct timer_list \*timer, unsigned long expires)":https://elixir.bootlin.com/linux/latest/source/include/linux/timer.h#L157\
* "int del_timer_sync(struct timer_list \*timer)": https://elixir.bootlin.com/linux/latest/source/include/linux/timer.h#L183

og en timer callback funktion på formen:

* ```static void my_timer_callback(struct timer_list *t)```

Toggler funktionaliten skal implementeres vha en kernel timer. Se bog og slides for detaljer herom. N.B. For at stoppe/slette en timer kaldes ```del_timer_sync(&timer)```.

Det er muligt at lave lave toggler funktionaliteten på to måder:

* Toggler funktionaliteten virker på alle gpio'er på en gang.
* Toggler funktionaliteten virker individuelt på de enkelte gpio'er, dvs. delay og state gælder per gpio. (Mere avanceret)

I begge tilfælde har attributterne tilhørende show og store metoder. Det følgende er lidt hjælp til implementeringen af disse.

#### Show metoden

Denne metode skal blot udlæse tilstanden af den pågældende attribute. Bruger man globale variabler for hhv. *state* og *delay* kan man nøjes med at læse værdien af den respektive tilbage og se bort fra det følgende omkring drvdata.

Har man *state* og *delay* for hver GPIO skal man have disse data tilknyttet det pågældende device (som er oprettet med device \_create...). Hvert oprettet device er af typen struct device, og denne indeholder et element, drvdata af typen void\*.\ Når man opretter et device med device_create kan man initialisere denne pointer med data som tilhører det pågældende device. Disse data kan siden tilgås i hhv. show()/store() funktionerne vha funktionen *dev_get_data()*. I det følgende eksempel initialiseres drvdata med en pointer til en int med gpio nummeret og siden hentes gpio nummeret ud i *show()* funktionen:

```C
static int gpio_no = 164; /* Global variabel with GPIO number */

chr_drv_probe(..) {
    ...
    device_create(my_class, NULL, MKDEV(MAJOR(devno), its_minor),
                &gpio_no, "mygpiodevicelabel");
...}

static ssize_t gpio_toggle_delay_show(struct device *dev, struct device_attribute *attr, char *buf) {

    int *gpio_in_use_ptr = dev_get_drvdata(dev);
    printk("Show accessed for device with GPIO %i\n", *gpio_in_use_ptr);
    ...
}
```

Skal man bruge dette i vores sammenhæng, ville det være smart at kunne få adgang til det rigtige element i vores gpio_devs struct.... ;-)

Desuden kunne det være smart at udvide vores gpio_devs struct til også at indeholde togge_state, toggle_delay og evt. en struct timer. Herved kan vi gemme toggle data for hver gpio port.

#### Store metoden

Store metoden er meget lig det vi kender fra *write* metoden. Vi behøver dog ikke kopiere fra user space. Ex:

```C
static ssize_t gpio_toggle_delay_store(struct device *dev,
        struct device_attribute *attr, const char *buf, size_t size)
{
    ssize_t ret = -EINVAL;
    unsigned long value;

    // Force zero termination to end of string
    buf[size-1] = 0; 

    // Convert string to unsigned long (s to ul)
    if((ret = kstrtoul(buf, 10, &value))<0)
    return ret;

    /* Do something with value */

    ret = size; // Always read the full content of buf
    return ret;
}
```

I den simple udgave skal vi blot opdatere de respektive globale variable (hhv. toggle_delay og toggle_state) og ved *toggle_state_store()* skal vi desuden initialisere timeren og starte den, hvis værdien er '1' eller delete den (del_timer_sync()), hvis værdien er '0'.

I udgaven med en timer per gpio, er det data'ene som kan tilgås via dev_get_drvdata() som skal opdateres, og timeren som skal startes. Bemærk!! timerens .data element kan indeholde en unsigned long int. Dette er netop hvad en pointer fylder, og man kan derfor angive en pointer, som kan derefereres og anvendes i timerens callback (handler) funktion.

**NOTE!** Det er vigtigt at få lukket timeren ned hvis modulet tages ud eller device tree unloades. Timeren lukkes ned med hjælp af *del_timer_sync()* find funktionsprototypen mm. med en søgning på <https://elixir.bootlin.com> bemærk at der i resultatet både hvor funktionen er defineret som prototype, hvor den er defineret som funktion (kode) og hvor man kan finde dokumentation. Hvad gør funktionen ud over at deaktivere timeren?

### Implementering af Timerfunktion

I alle tilfælde opdateres expiration time og gpio toggles.

I den simple udgave kan man vælge enten blot at toggle en enkelt gpio eller alle gpioer. Man kan også løbe alle gpio'er igennem og kun toggle de, for hvem "toggle_state" er sat til '1'.

Benytter man en timer per gpio, caster man unsigned long parameteren til en pointer af den rigtige struct type og tilgår via denne device's gpio nummer, timer og delay.


### Implementering af toggle, delay og timer funktioner mm.:

I denne opgave blev en platformdriver udvidet til at håndtere GPIO-toggling ved hjælp af timerfunktioner. Fokus var på at give hver GPIO en individuel toggling tilstand og forsinkelse, så de kunne styres uafhængigt af hinanden. Dette blev opnået ved at bruge en timer for hver GPIO, som kontrollerede toggling af tilstanden.

Den overordnede struktur består af flere komponenter:

GPIO-enhedens Struktur (gpio_dev): En struktur blev oprettet for at holde styr på oplysninger for hver GPIO-enhed. Denne struktur indeholder:
- gpio_number: Identifikationen af GPIO-porten.
- toggle_state: En variabel, der angiver, om GPIO'en er aktiv (1) eller inaktiv (0).
- toggle_delay: Den forsinkelse, der bruges ved toggling.
- timer: En struktur, der repræsenterer timeren for GPIO'en.

Strukturen gør det muligt at holde styr på hver enkelt GPIO’s tilstand og forsinkelse.
```c
struct gpio_dev {
    int no;   // GPIO nummer
    int dir;  // Retning
    int toggle_state;  // 1=toggling, 0=ikke toggling
    int toggle_delay;  // Delay i ms
    struct timer_list timer;  // Kernel-timer
};
```

Timer Callback Funktion: Timeren blev implementeret ved at bruge en callback-funktion, som bliver kaldt, når timeren udløber. Denne funktion toggler GPIO'ens tilstand, og hvis toggle_state er sat til 1, starter den timeren for at fortsætte togglingen.

Funktionens operation:
- Først hentes den relevante gpio_dev struktur ved hjælp af from_timer().
- Hvis toggle_state er 1 (aktivt), toggles GPIO’ens tilstand, og timeren opdateres med den forsinkelse, der er angivet i toggle_delay.
- Hvis toggle_state er 0, stopper toggling af GPIO'en

```c
static void gpio_timer_callback(struct timer_list *t) {
    struct gpio_dev *dev = from_timer(dev, t, timer);
    if (dev->toggle_state) {
        gpio_set_value(dev->no, !gpio_get_value(dev->no));
        mod_timer(&dev->timer, jiffies + msecs_to_jiffies(dev->toggle_delay));
    }
}
```
Show og Store Funktioner:
Sysfs-attributter blev brugt til at styre tilstanden og forsinkelsen for hver GPIO. For hver GPIO er der oprettet show og store funktioner for at kunne læse og skrive til toggle_state og toggle_delay.

- gpio_toggle_delay_show(): Denne funktion returnerer den aktuelle forsinkelse (delay) mellem toggling for en given GPIO. Værdien returneres som en streng, som kan læses af brugerland.
- gpio_toggle_delay_store(): Denne funktion opdaterer toggle_delay og justerer forsinkelsen mellem toggling. Når værdien ændres, justeres timerens expiration tid for at reflektere den nye forsinkelse.

Når toggle_delay opdateres, sørger store-funktionen for at justere timerens timing for at matche den nye forsinkelse.

- gpio_toggle_state_show(): Denne funktion returnerer den aktuelle tilstand af toggling (1 for aktiveret, 0 for deaktiveret) for en given GPIO.
- gpio_toggle_state_store(): Denne funktion opdaterer toggle_state og starter eller stopper timeren afhængigt af den værdi, der skrives til sysfs. Hvis toggle_state sættes til 1, startes timeren, mens den stoppes, hvis den sættes til 0.

Når toggle_state opdateres, sørger store-funktionen for enten at starte eller stoppe timeren baseret på den nye værdi.
```c
static ssize_t gpio_toggle_state_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    if (!gpio_dev) return -EINVAL;

    return sprintf(buf, "%d\n", gpio_dev->toggle_state);
}

static ssize_t gpio_toggle_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    int state;

    if (!gpio_dev) return -EINVAL;

    if (kstrtoint(buf, 10, &state)) return -EINVAL;

    gpio_dev->toggle_state = state;

    if (state) {
        mod_timer(&gpio_dev->timer, jiffies + msecs_to_jiffies(gpio_dev->toggle_delay));
    } else {
        del_timer_sync(&gpio_dev->timer);
    }

    return size;
}

static ssize_t gpio_toggle_delay_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    if (!gpio_dev) return -EINVAL;

    return sprintf(buf, "%d\n", gpio_dev->toggle_delay);
}

static ssize_t gpio_toggle_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    int delay;

    if (!gpio_dev) return -EINVAL;

    if (kstrtoint(buf, 10, &delay)) return -EINVAL;

    gpio_dev->toggle_delay = delay;
    return size;
}
```

Initialization af GPIO Devices: Når driveren initialiseres, oprettes der en GPIO-enhed for hver GPIO, som driveren skal håndtere. Hver enhed får tildelt en gpio_dev struktur, som knyttes til enheden via dev_set_drvdata(). Dette gør det muligt at tilgå de relevante data i show og store funktionerne samt i timer callbacken.

```c
DEVICE_ATTR_RW( gpio_toggle_delay );
DEVICE_ATTR_RW( gpio_toggle_state );

// Attributter til SysFS:
static struct attribute *gpio_attrs[] = {
    &dev_attr_gpio_toggle_state.attr,
    &dev_attr_gpio_toggle_delay.attr,
    NULL,
};
ATTRIBUTE_GROUPS( gpio );
```


### Test af Toggler på Target

Gør som i tidligere tests. Anvend echo til at sætte et toggle delay og starte din toggler på target. Check om LED'en lyser. Lyser den svagt kan det være fordi at den blinker hurtigt. Anvend evt printk/dmesg beskeder til at debugge og verificere dit arbejde. Check desuden gerne med Analog Discovery hvor godt toggle frekvensen passer, hvor meget [jitter](https://e2e.ti.com/blogs_/archives/b/energyzarr/archive/2012/10/26/the-truth-about-jitter) der er på den og om den ændrer sig over tid.

Først overføres .ko- og .dtbo-filerne til Raspberry Pi'en ved hjælp af scp, som beskrevet i den tidligere del. Efter filerne er overført, indlæses .ko-modulet med kommandoen insmod, og .dtbo-overlayen aktiveres med dtoverlay -d. Derefter anvendes echo-kommandoen til at konfigurere både toggle_delay og toggle_state. For at tænde LED'en med en forsinkelse og derefter slukke den, kan echo og cat anvendes som følger:

![dmesg_X2](dmesg_X2.png)

![catecho_x2](catecho_x2.png)

Implementeringen af GPIO-toggling ved hjælp af en kernel-timer blev gennemført med succes, og de ønskede funktioner blev opnået. Systemet fungerer som forventet, og det er nu muligt at tænde og slukke for toggling ved at skrive værdien 1 eller 0 til gpio_toggle_state via sysfs og echo. Hastigheden på togglingen kan også justeres ved at skrive en forsinkelsesværdi i millisekunder til gpio_toggle_delay. Den aktuelle tilstand af både gpio_toggle_state og gpio_toggle_delay kan observeres ved hjælp af cat.

